/* Copyright (C) 2013 by Daniel Stenberg et al
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

#include "CuTest.h"
#include "ares.h"
#include "test_utils.h"
#include <string.h>

/*
 * The TXT record all those tests are based on is coming from a tcpdump
 * done while requesting this record from bind9:
 * first._medianet-nms._tcp.example.com. TXT (
 *       "url=https://localhost/"
 *       "medianet-flows"
 *       "medianet-thresholds=https://localhost/thresholds"
 *       "period=1")
 */

static struct ares_txt_reply* txt_equals(CuTest* tc, const char* expected,
                                         const struct ares_txt_reply* txt_out)
  {
    CuAssertPtrNotNull(tc, txt_out);
    CuAssertStrEquals(tc, expected, (const char*)txt_out->txt);
    CuAssertIntEquals(tc, strlen(expected), txt_out->length);
    return txt_out->next;
  }

static void test_parse_txt_reply(CuTest* tc)
  {
    struct ares_txt_reply* txt_out;
    const struct ares_txt_reply* cur;
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e10"
        "0060"  /* RR length */

        /* url=https://localhost/ */
        "1675726c3d68747470733a2f2f6c6f63616c686f73742f"

        /* medianet-flows */
        "0e6d656469616e65742d666c6f7773"

        /* medianet-thresholds=https://localhost/thresholds */
        "306d656469616e65742d74687265"
        "73686f6c64733d68747470733a2f2f6c"
        "6f63616c686f73742f7468726573686f"
        "6c6473"

        /* period=1 */
        "08706572696f643d31"

        "c0250002"
        "000100000e10000b096c6f63616c686f"
        "737400c0ae0001000100093a8000047f"
        "000001c0ae001c000100093a80001000"
        "000000000000000000000000000001";
    struct frame udp_frame = hex2frame(hex_udp_frame);
    CuAssertIntEquals(tc, ARES_SUCCESS, ares_parse_txt_reply(udp_frame.abuf,
        udp_frame.alen, &txt_out));
    cur = txt_out;
    cur = txt_equals(tc, "url=https://localhost/", cur);
    cur = txt_equals(tc, "medianet-flows", cur);
    cur = txt_equals(tc, "medianet-thresholds=https://localhost/thresholds",
                     cur);
    cur = txt_equals(tc, "period=1", cur);
    CuAssertPtrEquals(tc, NULL, cur);
    free_frame(udp_frame);
    ares_free_data (txt_out);
  }

static void test_parse_txt_reply_bad_rr_len(CuTest* tc)
  {
    struct ares_txt_reply* txt_out;
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e10"
        "00FF"  /* RR length (bad value) */

        /* url=https://localhost/ */
        "1675726c3d68747470733a2f2f6c6f63616c686f73742f"

        /* medianet-flows */
        "0e6d656469616e65742d666c6f7773"

        /* medianet-thresholds=https://localhost/thresholds */
        "306d656469616e65742d74687265"
        "73686f6c64733d68747470733a2f2f6c"
        "6f63616c686f73742f7468726573686f"
        "6c6473"

        /* period=1 */
        "08706572696f643d31"

        "c0250002"
        "000100000e10000b096c6f63616c686f"
        "737400c0ae0001000100093a8000047f"
        "000001c0ae001c000100093a80001000"
        "000000000000000000000000000001";
    struct frame udp_frame = hex2frame(hex_udp_frame);
    CuAssertIntEquals(tc, ARES_EBADRESP, ares_parse_txt_reply(udp_frame.abuf,
        udp_frame.alen, &txt_out));
    CuAssertPtrEquals(tc, NULL, txt_out);
    free_frame(udp_frame);
  }

static void test_parse_txt_reply_bad_rr_header(CuTest* tc)
  {
    struct ares_txt_reply* txt_out;
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e";  /* huge cut, here */
    struct frame udp_frame = hex2frame(hex_udp_frame);
    CuAssertIntEquals(tc, ARES_EBADRESP, ares_parse_txt_reply(udp_frame.abuf,
        udp_frame.alen, &txt_out));
    CuAssertPtrEquals(tc, NULL, txt_out);
    free_frame(udp_frame);
  }

static void test_parse_txt_reply_bad_sub(CuTest* tc)
  {
    struct ares_txt_reply* txt_out;
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e10"
        "0060"  /* RR length */

        /* url=https://localhost/ */
        "1675726c3d68747470733a2f2f6c6f63616c686f73742f"

        /* medianet-flows */
        "0e6d656469616e65742d666c6f7773"

        /* medianet-thresholds=https://localhost/thresholds */
        "306d656469616e65742d74687265"
        "73686f6c64733d68747470733a2f2f6c"
        "6f63616c686f73742f7468726573686f"
        "6c6473"

        /* period=1 (with bad length) */
        "FF706572696f643d31"

        "c0250002"
        "000100000e10000b096c6f63616c686f"
        "737400c0ae0001000100093a8000047f"
        "000001c0ae001c000100093a80001000"
        "000000000000000000000000000001";
    struct frame udp_frame = hex2frame(hex_udp_frame);
    CuAssertIntEquals(tc, ARES_EBADRESP, ares_parse_txt_reply(udp_frame.abuf,
        udp_frame.alen, &txt_out));
    CuAssertPtrEquals(tc, NULL, txt_out);
    free_frame(udp_frame);
  }

static void test_parse_txt_reply_fuzzed(CuTest* tc)
  {
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e10"
        "0060"  /* RR length */

        /* url=https://localhost/ */
        "1675726c3d68747470733a2f2f6c6f63616c686f73742f"

        /* medianet-flows */
        "0e6d656469616e65742d666c6f7773"

        /* medianet-thresholds=https://localhost/thresholds */
        "306d656469616e65742d74687265"
        "73686f6c64733d68747470733a2f2f6c"
        "6f63616c686f73742f7468726573686f"
        "6c6473"

        /* period=1 */
        "08706572696f643d31"

        "c0250002"
        "000100000e10000b096c6f63616c686f"
        "737400c0ae0001000100093a8000047f"
        "000001c0ae001c000100093a80001000"
        "000000000000000000000000000001";
    struct frame udp_frame = hex2frame(hex_udp_frame);
    size_t pos;

    for (pos = 0; pos < udp_frame.alen; ++pos) {
        struct ares_txt_reply* txt_out;
        unsigned char orig = udp_frame.abuf[pos];
        int cur;

        for (cur = 0; cur <= 0xFF; ++cur) {
            udp_frame.abuf[pos] = cur;
            if (ares_parse_txt_reply(udp_frame.abuf, udp_frame.alen, &txt_out)
                == ARES_SUCCESS) {
                ares_free_data (txt_out);
            } else {
                CuAssertPtrEquals(tc, NULL, txt_out);
            }
        }
        udp_frame.abuf[pos] = orig;
    }

    free_frame(udp_frame);
  }

static void test_parse_txt_reply_cut(CuTest* tc)
  {
    const char* hex_udp_frame =
        "accf85800001"
        "0001000100020566697273740d5f6d65"
        "6469616e65742d6e6d73045f74637007"
        "6578616d706c6503636f6d0000100001"
        "c00c0010000100000e10"
        "0060"  /* RR length */

        /* url=https://localhost/ */
        "1675726c3d68747470733a2f2f6c6f63616c686f73742f"

        /* medianet-flows */
        "0e6d656469616e65742d666c6f7773"

        /* medianet-thresholds=https://localhost/thresholds */
        "306d656469616e65742d74687265"
        "73686f6c64733d68747470733a2f2f6c"
        "6f63616c686f73742f7468726573686f"
        "6c6473"

        /* period=1 */
        "08706572696f643d31"

        "c0250002"
        "000100000e10000b096c6f63616c686f"
        "737400c0ae0001000100093a8000047f"
        "000001c0ae001c000100093a80001000"
        "000000000000000000000000000001";
    struct frame udp_frame = hex2frame(hex_udp_frame);

    for (; udp_frame.alen>0; --udp_frame.alen) {
        struct ares_txt_reply* txt_out;
        if (ares_parse_txt_reply(udp_frame.abuf, udp_frame.alen, &txt_out)
            == ARES_SUCCESS) {
            ares_free_data (txt_out);
        } else {
            CuAssertPtrEquals(tc, NULL, txt_out);
        }
    }

    free_frame(udp_frame);
  }

CuSuite* create_parse_txt_suite()
  {
    CuSuite* suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, test_parse_txt_reply);
    SUITE_ADD_TEST(suite, test_parse_txt_reply_bad_rr_len);
    SUITE_ADD_TEST(suite, test_parse_txt_reply_bad_rr_header);
    SUITE_ADD_TEST(suite, test_parse_txt_reply_bad_sub);

    /* those are just to check if there is no crash (better results if run with
       valgrind --leak-check=full) */
    SUITE_ADD_TEST(suite, test_parse_txt_reply_fuzzed);
    SUITE_ADD_TEST(suite, test_parse_txt_reply_cut);
    return suite;
  }
